//
// SIFConvert.cpp : Take all of the Andor .SIF files in a given directory and
//                  convert them to ASCII format
//
// Author: Stephen Granade                Date: 23 Mar 2001
//

#include "stdafx.h"

#define ASCIIDIR   "ASCII Data"

// Prototypes
BOOL processSIFFile(const char *lpFilename);
BOOL readOneLine(HANDLE hFile, char *lpBuffer);
void printWinError(void);
void pause(void);

int main(int argc, char* argv[])
{
	char			lpDirectory[MAX_PATH];
	WIN32_FIND_DATA	wfd;
	HANDLE			hSearch;

	// Before we do anything else, make sure that the float datatype is exactly 4 bytes
	// (c.f. processSIFFile() for why this is important)
	if (sizeof(float) != 4) {
		printf("URGENT! The float data type is not 4 bytes long! Program no worky!\n");
		pause();
		return 1;
	}

	// Get the directory we'll be working with. If it's passed to us on the command
	// line, convert the 8.3 MSDOS-format path to the full long path
	if (argc > 1) {
		GetLongPathName(argv[1], lpDirectory, MAX_PATH);
	}
	else {
		printf("Directory to convert: ");
		gets(lpDirectory);
	}

	// Make sure we can go to that directory
	if (!SetCurrentDirectory(lpDirectory)) {
		printWinError();
		pause();
		return 1;
	}

	// Create the directory to hold the converted ASCII data
	CreateDirectory(ASCIIDIR, NULL);
	
	// Start processing all SIF files
	hSearch = FindFirstFile("*.sif", &wfd);
	if (hSearch == INVALID_HANDLE_VALUE) {
		printf("No .sif files found in that directory.\n");
		pause();
		return 1;
	}
	processSIFFile(wfd.cFileName);

	while (FindNextFile(hSearch, &wfd)) {
		processSIFFile(wfd.cFileName);
	}

	FindClose(hSearch);
	pause();
	return 0;
}

// Process a given SIF file. SIF files consist of header lines separated by character
// 0x0A (aka '\n') and then a block of binary data. Data is stored using the IEEE-754
// standard for 32-bit single-precision floating numbers. The dimensions of the data
// are found on line 5 of the header, and there are 22 total lines in the header
BOOL processSIFFile(const char *lpFilename)
{
	HANDLE	hInFile, hOutFile;
	int		iDimensionLine = 5,		// Where the (x, y) data size info is
			iHeaderLines = 22,
			iX, iY,					// The dimensions of the data
			i, j;
	char	lpBuffer[16535],		// A scratch buffer
			lpBuf2[4096],			// Another scratch buffer
			lpNewFile[MAX_PATH],    // The new data filename
			*pc;
	float   *pfData;                // The data from the SIF file
	DWORD	dwBytes;
	BOOL	rc = 1;					// Return code

	printf("Processing file \"%s\"", lpFilename);

	// Create the filename for the ASCII data
	strcpy(lpNewFile, ASCIIDIR);
	strcat(lpNewFile, "\\");
	strcat(lpNewFile, lpFilename);
	if (!(pc = strstr(lpNewFile, ".sif"))) {
		printf(" -- the file must end in '.sif'\n");
		return 0;
	}
	strcpy(pc, ".txt");

	// Open the .SIF file
	hInFile = CreateFile(lpFilename, GENERIC_READ, 0, NULL,
		OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if (hInFile == INVALID_HANDLE_VALUE) {
	    printWinError();
		return 0;
	}
	
	// Deal with the header information
	for (i = 1; i <= iHeaderLines; i++) {
		readOneLine(hInFile, lpBuffer);
		if (i == iDimensionLine) {
			if (sscanf(lpBuffer, " %i %i", &iX, &iY) != 2) {
				printf(" -- unable to read dimension data\n");
				CloseHandle(hInFile);
				return 0;
			}
			printf(" (%i, %i)", iX, iY);
		}
	}

	// At this point the file pointer is pointing square at the binary data.
	// A word about converting the data. Each data point is stored as a single-
	// precision 4 byte floating-point number (IEEE-754 standard). This standard
	// is the same as is used by most every C compiler. I'm going to assume that
	// the compiler uses the same IEEE-754 format as the SIF file does, and
	// more importantly, that the endian-ness is the same. SIF data is stored
	// little-endian first. This is *very important* if you want to port this
	// program, or even if you want to understand what the hell I'm doing.
	//
	// So what I'm going to do is create an array of floats the proper size, then
	// pretend that that array is a byte buffer and read the data into it,
	//
	// Yeah, I know. This is a nasty kludge.
	pfData = (float *)malloc((size_t)(iX * iY * sizeof(float)));

	if (!ReadFile(hInFile, (char *)pfData, (iX * iY * sizeof(float)), &dwBytes, NULL)) {
		printWinError();
		rc = 0;
	}
	else if (dwBytes != iX * iY * sizeof(float)) {
		printf("\n\tUnable to read in all the data (%i)", dwBytes);
		rc = 0;
	}
	else if ((hOutFile = CreateFile(lpNewFile, GENERIC_WRITE, 0, NULL,
		CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE) {
	    printWinError();
		rc = 0;
	}
	else {
		// Data printing time!
		for (i = 0; i < iX; i++) {
			sprintf(lpBuffer, "%i", i+1);
			for (j = 0; j < iY; j++) {
				sprintf(lpBuf2, "\t%.6g", pfData[i + j*iX]);
				strcat(lpBuffer,lpBuf2);
			}
			strcat(lpBuffer, "\n");
			if (!WriteFile(hOutFile, lpBuffer, strlen(lpBuffer), &dwBytes, NULL)) {
				printf("\n");
				printWinError();
				rc = 0;
				break;
			}
		}				
		CloseHandle(hOutFile);
	}

	printf("\n");

	free(pfData);
	CloseHandle(hOutFile);
	CloseHandle(hInFile);

	return rc;
}

// Read one line from a file, assuming that the line ends in '\n'
BOOL readOneLine(HANDLE hFile, char *lpBuffer)
{
	int		i=0;
	DWORD	dwBytesRead;
	char	c;
	
	while (1) {
		if (!ReadFile(hFile, &c, 1, &dwBytesRead, NULL)) {
			printWinError();
			return 0;
		}
		if (c == '\n') {
			lpBuffer[i] = NULL;
			return 1;
		}
		lpBuffer[i++] = c;
	}
}

// Print any error message that Windows reports
void printWinError(void)
{
	char lpBuffer[4096];

	FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL, GetLastError(),
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
		(LPTSTR) lpBuffer, 4092, NULL);
	printf("Error: %s", lpBuffer);
	LocalFree(lpBuffer);
}

// Pause for a keypress, assuming bP is true
void pause(void)
{
	printf("\nPress a key to end the program\n");
	_getch();
}
